import gleam/bit_array
import gleam/dict.{type Dict}
import gleam/dynamic/decode.{type Decoder}
import gleam/http.{type Method}
import gleam/http/request.{type Request, Request}
import gleam/http/response.{type Response}
import gleam/int
import gleam/json.{type DecodeError, type Json}
import gleam/list
import gleam/option.{type Option}
import gleam/string_tree
import gleam/uri.{type Uri}
import spacetraders_models/account.{type Account}
import spacetraders_models/account_token.{type AccountToken}
import spacetraders_models/agent.{type Agent}
import spacetraders_models/agent_event.{type AgentEvent}
import spacetraders_models/agent_symbol.{type AgentSymbol}
import spacetraders_models/agent_token.{type AgentToken}
import spacetraders_models/auth_method.{
  type AuthMethod, AccountAuth, AgentAuth, NoAuth,
}
import spacetraders_models/chart.{type Chart}
import spacetraders_models/chart_transaction.{type ChartTransaction}
import spacetraders_models/construction.{type Construction}
import spacetraders_models/contract.{type Contract}
import spacetraders_models/contract_id.{type ContractId}
import spacetraders_models/cooldown.{type Cooldown}
import spacetraders_models/date.{type Date}
import spacetraders_models/extraction.{type Extraction}
import spacetraders_models/faction.{type Faction}
import spacetraders_models/faction_symbol.{type FactionSymbol}
import spacetraders_models/jump_gate.{type JumpGate}
import spacetraders_models/market.{type Market}
import spacetraders_models/market_transaction.{type MarketTransaction}
import spacetraders_models/module_symbol.{type ModuleSymbol}
import spacetraders_models/mount_symbol.{type MountSymbol}
import spacetraders_models/paged_data.{type PagedData}
import spacetraders_models/public_agent.{type PublicAgent}
import spacetraders_models/refinement_produce.{type RefinementProduce}
import spacetraders_models/refinement_yield.{type RefinementYield}
import spacetraders_models/repair_transaction.{type RepairTransaction}
import spacetraders_models/scanned_ship.{type ScannedShip}
import spacetraders_models/scanned_system.{type ScannedSystem}
import spacetraders_models/scanned_waypoint.{type ScannedWaypoint}
import spacetraders_models/scrap_transaction.{type ScrapTransaction}
import spacetraders_models/ship.{type Ship}
import spacetraders_models/ship_cargo.{type ShipCargo}
import spacetraders_models/ship_condition_event.{type ShipConditionEvent}
import spacetraders_models/ship_fuel.{type ShipFuel}
import spacetraders_models/ship_modification_transaction.{
  type ShipModificationTransaction,
}
import spacetraders_models/ship_module.{type ShipModule}
import spacetraders_models/ship_mount.{type ShipMount}
import spacetraders_models/ship_nav.{type ShipNav}
import spacetraders_models/ship_nav_flight_mode.{type ShipNavFlightMode}
import spacetraders_models/ship_symbol.{type ShipSymbol}
import spacetraders_models/ship_type.{type ShipType}
import spacetraders_models/shipyard.{type Shipyard}
import spacetraders_models/shipyard_transaction.{type ShipyardTransaction}
import spacetraders_models/siphon.{type Siphon}
import spacetraders_models/spacetraders_error.{type SpacetradersError}
import spacetraders_models/survey.{type Survey}
import spacetraders_models/system.{type System}
import spacetraders_models/system_symbol.{type SystemSymbol}
import spacetraders_models/timestamp.{type Timestamp}
import spacetraders_models/trade_symbol.{type TradeSymbol}
import spacetraders_models/waypoint.{type Waypoint}
import spacetraders_models/waypoint_modifier.{type WaypointModifier}
import spacetraders_models/waypoint_symbol.{type WaypointSymbol}
import spacetraders_models/waypoint_trait_symbol.{type WaypointTraitSymbol}
import spacetraders_models/waypoint_type.{type WaypointType}

pub type ApiError {
  JsonDecodeError(DecodeError)
  ServerError(SpacetradersError)
}

pub type ApiResponse(data) =
  Result(data, ApiError)

const base_request = Request(
  method: http.Get,
  headers: [],
  body: <<>>,
  scheme: http.Https,
  host: "api.spacetraders.io",
  port: option.None,
  path: "/v2",
  query: option.None,
)

fn set_body(
  req: Request(a),
  content_type: String,
  body: BitArray,
) -> Request(BitArray) {
  req
  |> request.set_header("Content-Type", content_type)
  |> request.set_body(body)
}

fn set_json_body(req: Request(a), json: Json) {
  req
  |> set_body(
    "application/json",
    json
      |> json.to_string_tree
      |> string_tree.to_string
      |> bit_array.from_string,
  )
}

fn make_request(method: Method, auth_method: AuthMethod, path: String) {
  base_request
  |> request.set_method(method)
  |> auth_method.apply(auth_method)
  |> request.set_path(base_request.path <> path)
}

fn get(auth_method: AuthMethod, path: String) {
  make_request(http.Get, auth_method, path)
}

fn page_query_params(
  initial_query: List(#(String, String)),
  page: Option(Int),
  limit: Option(Int),
) -> List(#(String, String)) {
  let query = initial_query
  let query = case page {
    option.Some(page) -> [#("page", int.to_string(page)), ..query]
    option.None -> query
  }
  let query = case limit {
    option.Some(page) -> [#("limit", int.to_string(page)), ..query]
    option.None -> query
  }
  query
}

fn get_page(
  auth_method: AuthMethod,
  path: String,
  page: Option(Int),
  limit: Option(Int),
) {
  make_request(http.Get, auth_method, path)
  |> request.set_query(page_query_params([], page, limit))
}

fn get_page_with_query(
  auth_method: AuthMethod,
  path: String,
  page: Option(Int),
  limit: Option(Int),
  query: List(#(String, String)),
) -> Request(BitArray) {
  make_request(http.Get, auth_method, path)
  |> request.set_query(page_query_params(query, page, limit))
}

fn post(auth_method: AuthMethod, path: String) -> Request(BitArray) {
  make_request(http.Post, auth_method, path)
}

fn post_json(
  auth_method: AuthMethod,
  path: String,
  body: Json,
) -> Request(BitArray) {
  make_request(http.Post, auth_method, path) |> set_json_body(body)
}

fn patch_json(
  auth_method: AuthMethod,
  path: String,
  body: Json,
) -> Request(BitArray) {
  make_request(http.Patch, auth_method, path) |> set_json_body(body)
}

fn parse_response(
  response: Response(BitArray),
  decoder: Decoder(data),
) -> ApiResponse(data) {
  case json.parse_bits(response.body, decoder) {
    Ok(data) -> Ok(data)
    Error(err) -> Error(JsonDecodeError(err))
  }
}

fn data_decoder(decoder: Decoder(data)) -> Decoder(data) {
  use data <- decode.field("data", decoder)
  decode.success(data)
}

fn parse_data_response(
  response: Response(BitArray),
  decoder: Decoder(data),
) -> ApiResponse(data) {
  parse_response(response, data_decoder(decoder))
}

fn parse_paged_data_response(
  response: Response(BitArray),
  decoder: Decoder(data),
) -> ApiResponse(PagedData(data)) {
  parse_response(response, paged_data.decoder(decoder))
}

fn error_decoder() -> Decoder(SpacetradersError) {
  use data <- decode.field("error", spacetraders_error.decoder())
  decode.success(data)
}

fn parse_error_response(response: Response(BitArray)) -> ApiResponse(data) {
  Error(case parse_response(response, error_decoder()) {
    Ok(err) -> ServerError(err)
    Error(err) -> err
  })
}

pub fn get_account_request(token: AgentToken) -> Request(BitArray) {
  get(AgentAuth(token), "/my/account")
}

pub fn get_account_response(
  response: Response(BitArray),
) -> ApiResponse(Account) {
  case response.status {
    200 ->
      parse_data_response(
        response,
        decode.field("account", account.decoder(), decode.success),
      )
    _ -> parse_error_response(response)
  }
}

pub type AgentRegistered {
  AgentRegistered(
    token: AgentToken,
    agent: Agent,
    faction: Faction,
    contract: Contract,
    ships: List(Ship),
  )
}

pub fn register_new_agent_request(
  token: AccountToken,
  agent_symbol: AgentSymbol,
  faction_symbol: FactionSymbol,
) -> Request(BitArray) {
  post_json(
    AccountAuth(token),
    "/register",
    json.object([
      #("symbol", agent_symbol.encode(agent_symbol)),
      #("faction", faction_symbol.encode(faction_symbol)),
    ]),
  )
}

pub fn register_new_agent_response(
  response: Response(BitArray),
) -> ApiResponse(AgentRegistered) {
  case response.status {
    201 ->
      parse_data_response(response, {
        use token <- decode.field("token", agent_token.decoder())
        use agent <- decode.field("agent", agent.decoder())
        use faction <- decode.field("faction", faction.decoder())
        use contract <- decode.field("contract", contract.decoder())
        use ships <- decode.field("ships", decode.list(ship.decoder()))
        decode.success(AgentRegistered(
          token:,
          agent:,
          faction:,
          contract:,
          ships:,
        ))
      })
    _ -> parse_error_response(response)
  }
}

pub fn list_public_agents_request(
  page: Option(Int),
  limit: Option(Int),
) -> Request(BitArray) {
  get_page(NoAuth, "/agents", page, limit)
}

pub fn list_public_agents_response(
  response: Response(BitArray),
) -> ApiResponse(PagedData(List(PublicAgent))) {
  case response.status {
    200 ->
      parse_paged_data_response(response, decode.list(public_agent.decoder()))
    _ -> parse_error_response(response)
  }
}

pub fn get_public_agent_request(agent_symbol: AgentSymbol) -> Request(BitArray) {
  get(NoAuth, "/agents/" <> agent_symbol.to_string(agent_symbol))
}

pub fn get_public_agent_response(
  response: Response(BitArray),
) -> ApiResponse(PublicAgent) {
  case response.status {
    200 -> parse_data_response(response, public_agent.decoder())
    _ -> parse_error_response(response)
  }
}

pub fn get_agent_request(token: AgentToken) -> Request(BitArray) {
  get(AgentAuth(token), "/my/agent")
}

pub fn get_agent_response(response: Response(BitArray)) -> ApiResponse(Agent) {
  case response.status {
    200 -> parse_data_response(response, agent.decoder())
    _ -> parse_error_response(response)
  }
}

pub fn get_agent_events_request(token: AgentToken) -> Request(BitArray) {
  get(AgentAuth(token), "/my/agent/events")
}

pub fn get_agent_events_response(
  response: Response(BitArray),
) -> ApiResponse(List(AgentEvent)) {
  case response.status {
    200 -> parse_data_response(response, decode.list(agent_event.decoder()))
    _ -> parse_error_response(response)
  }
}

pub fn list_contracts_request(
  token: AgentToken,
  page: Option(Int),
  limit: Option(Int),
) -> Request(BitArray) {
  get_page(AgentAuth(token), "/my/contracts", page, limit)
}

pub fn list_contracts_response(
  response: Response(BitArray),
) -> ApiResponse(PagedData(List(Contract))) {
  case response.status {
    200 -> parse_paged_data_response(response, decode.list(contract.decoder()))
    _ -> parse_error_response(response)
  }
}

pub fn get_contract_request(
  token: AgentToken,
  contract_id: ContractId,
) -> Request(BitArray) {
  get(AgentAuth(token), "/my/contracts/" <> contract_id.to_string(contract_id))
}

pub fn get_contract_response(
  response: Response(BitArray),
) -> ApiResponse(Contract) {
  case response.status {
    200 -> parse_data_response(response, contract.decoder())
    _ -> parse_error_response(response)
  }
}

pub type ContractAccepted {
  ContractAccepted(contract: Contract, agent: Agent)
}

pub fn accept_contract_request(
  token: AgentToken,
  contract_id: ContractId,
) -> Request(BitArray) {
  post(
    AgentAuth(token),
    "/my/contracts/" <> contract_id.to_string(contract_id) <> "/accept",
  )
}

pub fn accept_contract_response(
  response: Response(BitArray),
) -> ApiResponse(ContractAccepted) {
  case response.status {
    200 ->
      parse_data_response(response, {
        use contract <- decode.field("contract", contract.decoder())
        use agent <- decode.field("agent", agent.decoder())
        decode.success(ContractAccepted(contract:, agent:))
      })
    _ -> parse_error_response(response)
  }
}

pub type ContractFulfilled {
  ContractFulfilled(contract: Contract, agent: Agent)
}

pub fn fulfill_contract_request(
  token: AgentToken,
  contract_id: ContractId,
) -> Request(BitArray) {
  post(
    AgentAuth(token),
    "/my/contracts/" <> contract_id.to_string(contract_id) <> "/fulfill",
  )
}

pub fn fulfill_contract_response(
  response: Response(BitArray),
) -> ApiResponse(ContractFulfilled) {
  case response.status {
    200 ->
      parse_data_response(response, {
        use contract <- decode.field("contract", contract.decoder())
        use agent <- decode.field("agent", agent.decoder())
        decode.success(ContractFulfilled(contract:, agent:))
      })
    _ -> parse_error_response(response)
  }
}

pub type ContractCargoDelivered {
  ContractCargoDelivered(contract: Contract, cargo: ShipCargo)
}

pub fn deliver_contract_cargo_request(
  token: AgentToken,
  contract_id: ContractId,
  ship_symbol: ShipSymbol,
  trade_symbol: TradeSymbol,
  units: Int,
) -> Request(BitArray) {
  post_json(
    AgentAuth(token),
    "/my/contracts/" <> contract_id.to_string(contract_id) <> "/deliver",
    json.object([
      #("shipSymbol", ship_symbol.encode(ship_symbol)),
      #("tradeSymbol", trade_symbol.encode(trade_symbol)),
      #("units", json.int(units)),
    ]),
  )
}

pub fn deliver_contract_cargo_response(
  response: Response(BitArray),
) -> ApiResponse(ContractCargoDelivered) {
  case response.status {
    200 ->
      parse_data_response(response, {
        use contract <- decode.field("contract", contract.decoder())
        use cargo <- decode.field("cargo", ship_cargo.decoder())
        decode.success(ContractCargoDelivered(contract:, cargo:))
      })
    _ -> parse_error_response(response)
  }
}

pub type ExportToImportMap =
  Dict(TradeSymbol, List(TradeSymbol))

pub fn get_supply_chain_request() -> Request(BitArray) {
  get(NoAuth, "/market/supply-chain")
}

pub fn get_supply_chain_response(
  response: Response(BitArray),
) -> ApiResponse(ExportToImportMap) {
  case response.status {
    200 ->
      parse_data_response(
        response,
        decode.field(
          "exportToImportMap",
          decode.dict(
            trade_symbol.decoder(),
            decode.list(trade_symbol.decoder()),
          ),
          decode.success,
        ),
      )
    _ -> parse_error_response(response)
  }
}

pub fn list_factions_request(
  page: Option(Int),
  limit: Option(Int),
) -> Request(BitArray) {
  get_page(NoAuth, "/factions", page, limit)
}

pub fn list_factions_response(
  response: Response(BitArray),
) -> ApiResponse(PagedData(List(Faction))) {
  case response.status {
    200 -> parse_paged_data_response(response, decode.list(faction.decoder()))
    _ -> parse_error_response(response)
  }
}

pub fn get_faction_request(symbol: FactionSymbol) -> Request(BitArray) {
  get(NoAuth, "/factions/" <> faction_symbol.to_string(symbol))
}

pub fn get_faction_response(
  response: Response(BitArray),
) -> ApiResponse(Faction) {
  case response.status {
    200 -> parse_data_response(response, faction.decoder())
    _ -> parse_error_response(response)
  }
}

pub type FactionReputation {
  FactionReputation(symbol: FactionSymbol, reputation: Int)
}

pub fn get_my_factions_request(
  token: AgentToken,
  page: Option(Int),
  limit: Option(Int),
) -> Request(BitArray) {
  get_page(AgentAuth(token), "/my/factions", page, limit)
}

pub fn get_my_factions_response(
  response: Response(BitArray),
) -> ApiResponse(PagedData(List(FactionReputation))) {
  case response.status {
    200 ->
      parse_paged_data_response(
        response,
        decode.list({
          use symbol <- decode.field("symbol", faction_symbol.decoder())
          use reputation <- decode.field("reputation", decode.int)
          decode.success(FactionReputation(symbol:, reputation:))
        }),
      )
    _ -> parse_error_response(response)
  }
}

pub fn list_ships_request(
  token: AgentToken,
  page: Option(Int),
  limit: Option(Int),
) -> Request(BitArray) {
  get_page(AgentAuth(token), "/my/ships", page, limit)
}

pub fn list_ships_response(
  response: Response(BitArray),
) -> ApiResponse(PagedData(List(Ship))) {
  case response.status {
    200 -> parse_paged_data_response(response, decode.list(ship.decoder()))
    _ -> parse_error_response(response)
  }
}

pub type ShipPurchased {
  ShipPurchased(ship: Ship, agent: Agent, transaction: ShipyardTransaction)
}

pub fn purchase_ship_request(
  token: AgentToken,
  ship_type: ShipType,
  waypoint_symbol: WaypointSymbol,
) -> Request(BitArray) {
  post_json(
    AgentAuth(token),
    "/my/ships",
    json.object([
      #("shipType", ship_type.encode(ship_type)),
      #("waypointSymbol", waypoint_symbol.encode(waypoint_symbol)),
    ]),
  )
}

pub fn purchase_ship_response(
  response: Response(BitArray),
) -> ApiResponse(ShipPurchased) {
  case response.status {
    201 ->
      parse_data_response(response, {
        use ship <- decode.field("ship", ship.decoder())
        use agent <- decode.field("agent", agent.decoder())
        use transaction <- decode.field(
          "transaction",
          shipyard_transaction.decoder(),
        )
        decode.success(ShipPurchased(ship:, agent:, transaction:))
      })
    _ -> parse_error_response(response)
  }
}

pub fn get_ship_request(
  token: AgentToken,
  ship_symbol: ShipSymbol,
) -> Request(BitArray) {
  get(AgentAuth(token), "/my/ships/" <> ship_symbol.to_string(ship_symbol))
}

pub fn get_ship_response(response: Response(BitArray)) -> ApiResponse(Ship) {
  case response.status {
    200 -> parse_data_response(response, ship.decoder())
    _ -> parse_error_response(response)
  }
}

pub type ChartCreated {
  ChartCreated(
    chart: Chart,
    waypoint: Waypoint,
    transaction: ChartTransaction,
    agent: Agent,
  )
}

pub fn create_chart_request(
  token: AgentToken,
  ship_symbol: ShipSymbol,
) -> Request(BitArray) {
  post(
    AgentAuth(token),
    "/my/ships/" <> ship_symbol.to_string(ship_symbol) <> "/chart",
  )
}

pub fn create_chart_response(
  response: Response(BitArray),
) -> ApiResponse(ChartCreated) {
  case response.status {
    201 ->
      parse_data_response(response, {
        use chart <- decode.field("chart", chart.decoder())
        use waypoint <- decode.field("waypoint", waypoint.decoder())
        use transaction <- decode.field(
          "transaction",
          chart_transaction.decoder(),
        )
        use agent <- decode.field("agent", agent.decoder())
        decode.success(ChartCreated(chart:, waypoint:, transaction:, agent:))
      })
    _ -> parse_error_response(response)
  }
}

pub fn negotiate_contract_request(
  token: AgentToken,
  ship_symbol: String,
) -> Request(BitArray) {
  post(AgentAuth(token), "/my/ships/" <> ship_symbol <> "/negotiate/contract")
}

pub fn negotiate_contract_response(
  response: Response(BitArray),
) -> ApiResponse(Contract) {
  case response.status {
    201 ->
      parse_data_response(
        response,
        decode.field("contract", contract.decoder(), decode.success),
      )
    _ -> parse_error_response(response)
  }
}

pub fn get_ship_cooldown_request(
  token: AgentToken,
  ship_symbol: ShipSymbol,
) -> Request(BitArray) {
  get(
    AgentAuth(token),
    "/my/ships/" <> ship_symbol.to_string(ship_symbol) <> "/cooldown",
  )
}

pub fn get_ship_cooldown_response(
  response: Response(BitArray),
) -> ApiResponse(Option(Cooldown)) {
  case response.status {
    200 ->
      parse_data_response(
        response,
        decode.field("cooldown", cooldown.decoder(), fn(cooldown) {
          decode.success(option.Some(cooldown))
        }),
      )
    204 -> Ok(option.None)
    _ -> parse_error_response(response)
  }
}

pub fn dock_ship_request(
  token: AgentToken,
  ship_symbol: ShipSymbol,
) -> Request(BitArray) {
  post(
    AgentAuth(token),
    "/my/ships/" <> ship_symbol.to_string(ship_symbol) <> "/dock",
  )
}

pub fn dock_ship_response(response: Response(BitArray)) -> ApiResponse(ShipNav) {
  case response.status {
    200 ->
      parse_data_response(
        response,
        decode.field("nav", ship_nav.decoder(), decode.success),
      )
    _ -> parse_error_response(response)
  }
}

pub type ResourcesExtracted {
  ResourcesExtracted(
    extraction: Extraction,
    cooldown: Cooldown,
    cargo: ShipCargo,
    modifiers: List(WaypointModifier),
    events: List(ShipConditionEvent),
  )
}

fn resources_extracted_decoder() -> Decoder(ResourcesExtracted) {
  use extraction <- decode.field("extraction", extraction.decoder())
  use cooldown <- decode.field("cooldown", cooldown.decoder())
  use cargo <- decode.field("cargo", ship_cargo.decoder())
  use modifiers <- decode.field(
    "modifiers",
    decode.list(waypoint_modifier.decoder()),
  )
  use events <- decode.field(
    "events",
    decode.list(ship_condition_event.decoder()),
  )
  decode.success(ResourcesExtracted(
    extraction:,
    cooldown:,
    cargo:,
    modifiers:,
    events:,
  ))
}

pub fn extract_resources_request(
  token: AgentToken,
  ship_symbol: ShipSymbol,
) -> Request(BitArray) {
  post(
    AgentAuth(token),
    "/my/ships/" <> ship_symbol.to_string(ship_symbol) <> "/extract",
  )
}

pub fn extract_resources_response(
  response: Response(BitArray),
) -> ApiResponse(ResourcesExtracted) {
  case response.status {
    201 -> parse_data_response(response, resources_extracted_decoder())
    _ -> parse_error_response(response)
  }
}

pub fn extract_resources_with_survey_request(
  token: AgentToken,
  ship_symbol: ShipSymbol,
  survey: Survey,
) -> Request(BitArray) {
  post_json(
    AgentAuth(token),
    "/my/ships/" <> ship_symbol.to_string(ship_symbol) <> "/extract/survey",
    survey.encode(survey),
  )
}

pub fn extract_resources_with_survey_response(
  response: Response(BitArray),
) -> ApiResponse(ResourcesExtracted) {
  case response.status {
    201 -> parse_data_response(response, resources_extracted_decoder())
    _ -> parse_error_response(response)
  }
}

pub fn jettison_cargo_request(
  token: AgentToken,
  ship_symbol: ShipSymbol,
  trade_symbol: TradeSymbol,
  units: Int,
) -> Request(BitArray) {
  post_json(
    AgentAuth(token),
    "/my/ships/" <> ship_symbol.to_string(ship_symbol) <> "/jettison",
    json.object([
      #("symbol", trade_symbol.encode(trade_symbol)),
      #("units", json.int(units)),
    ]),
  )
}

pub fn jettison_cargo_response(
  response: Response(BitArray),
) -> ApiResponse(ShipCargo) {
  case response.status {
    200 ->
      parse_data_response(
        response,
        decode.field("cargo", ship_cargo.decoder(), decode.success),
      )
    _ -> parse_error_response(response)
  }
}

pub type ShipJumped {
  ShipJumped(
    nav: ShipNav,
    cooldown: Cooldown,
    transaction: MarketTransaction,
    agent: Agent,
  )
}

pub fn jump_ship_request(
  token: AgentToken,
  ship_symbol: ShipSymbol,
  waypoint_symbol: WaypointSymbol,
) -> Request(BitArray) {
  post_json(
    AgentAuth(token),
    "/my/ships/" <> ship_symbol.to_string(ship_symbol) <> "/jump",
    json.object([#("waypointSymbol", waypoint_symbol.encode(waypoint_symbol))]),
  )
}

pub fn jump_ship_response(
  response: Response(BitArray),
) -> ApiResponse(ShipJumped) {
  case response.status {
    200 ->
      parse_data_response(response, {
        use nav <- decode.field("nav", ship_nav.decoder())
        use cooldown <- decode.field("cooldown", cooldown.decoder())
        use transaction <- decode.field(
          "transaction",
          market_transaction.decoder(),
        )
        use agent <- decode.field("agent", agent.decoder())
        decode.success(ShipJumped(nav:, cooldown:, transaction:, agent:))
      })
    _ -> parse_error_response(response)
  }
}

pub type Scan(data) {
  Scan(cooldown: Cooldown, data: List(data))
}

fn scan_decoder(field: String, decoder: Decoder(data)) -> Decoder(Scan(data)) {
  use cooldown <- decode.field("cooldown", cooldown.decoder())
  use data <- decode.field(field, decode.list(decoder))
  decode.success(Scan(cooldown:, data:))
}

pub fn scan_systems_request(
  token: AgentToken,
  ship_symbol: ShipSymbol,
) -> Request(BitArray) {
  post(
    AgentAuth(token),
    "/my/ships/" <> ship_symbol.to_string(ship_symbol) <> "/scan/systems",
  )
}

pub fn scan_systems_response(
  response: Response(BitArray),
) -> ApiResponse(Scan(ScannedSystem)) {
  case response.status {
    201 ->
      parse_data_response(
        response,
        scan_decoder("systems", scanned_system.decoder()),
      )
    _ -> parse_error_response(response)
  }
}

pub fn scan_waypoints_request(
  token: AgentToken,
  ship_symbol: ShipSymbol,
) -> Request(BitArray) {
  post(
    AgentAuth(token),
    "/my/ships/" <> ship_symbol.to_string(ship_symbol) <> "/scan/waypoints",
  )
}

pub fn scan_waypoints_response(
  response: Response(BitArray),
) -> ApiResponse(Scan(ScannedWaypoint)) {
  case response.status {
    201 ->
      parse_data_response(
        response,
        scan_decoder("waypoints", scanned_waypoint.decoder()),
      )
    _ -> parse_error_response(response)
  }
}

pub fn scan_ships_request(
  token: AgentToken,
  ship_symbol: ShipSymbol,
) -> Request(BitArray) {
  post(
    AgentAuth(token),
    "/my/ships/" <> ship_symbol.to_string(ship_symbol) <> "/scan/ships",
  )
}

pub fn scan_ships_response(
  response: Response(BitArray),
) -> ApiResponse(Scan(ScannedShip)) {
  case response.status {
    201 ->
      parse_data_response(
        response,
        scan_decoder("ships", scanned_ship.decoder()),
      )
    _ -> parse_error_response(response)
  }
}

pub type ShipScrapped {
  ShipScrapped(agent: Agent, transaction: ScrapTransaction)
}

pub fn scrap_ship_request(
  token: AgentToken,
  ship_symbol: ShipSymbol,
) -> Request(BitArray) {
  post(
    AgentAuth(token),
    "/my/ships/" <> ship_symbol.to_string(ship_symbol) <> "/scrap",
  )
}

pub fn scrap_ship_response(
  response: Response(BitArray),
) -> ApiResponse(ShipScrapped) {
  case response.status {
    200 ->
      parse_data_response(response, {
        use agent <- decode.field("agent", agent.decoder())
        use transaction <- decode.field(
          "transaction",
          scrap_transaction.decoder(),
        )
        decode.success(ShipScrapped(agent:, transaction:))
      })
    _ -> parse_error_response(response)
  }
}

pub fn get_scrap_ship_request(
  token: AgentToken,
  ship_symbol: ShipSymbol,
) -> Request(BitArray) {
  get(
    AgentAuth(token),
    "/my/ships/" <> ship_symbol.to_string(ship_symbol) <> "/scrap",
  )
}

pub fn get_scrap_ship_response(
  response: Response(BitArray),
) -> ApiResponse(ScrapTransaction) {
  case response.status {
    200 ->
      parse_data_response(
        response,
        decode.field("transaction", scrap_transaction.decoder(), decode.success),
      )
    _ -> parse_error_response(response)
  }
}

pub type ShipNavigated {
  ShipNavigated(nav: ShipNav, fuel: ShipFuel, events: List(ShipConditionEvent))
}

pub fn navigate_ship_request(
  token: AgentToken,
  ship_symbol: ShipSymbol,
  waypoint_symbol: WaypointSymbol,
) -> Request(BitArray) {
  post_json(
    AgentAuth(token),
    "/my/ships/" <> ship_symbol.to_string(ship_symbol) <> "/navigate",
    json.object([#("waypointSymbol", waypoint_symbol.encode(waypoint_symbol))]),
  )
}

pub fn navigate_ship_response(
  response: Response(BitArray),
) -> ApiResponse(ShipNavigated) {
  case response.status {
    200 ->
      parse_data_response(response, {
        use nav <- decode.field("nav", ship_nav.decoder())
        use fuel <- decode.field("fuel", ship_fuel.decoder())
        use events <- decode.field(
          "events",
          decode.list(ship_condition_event.decoder()),
        )
        decode.success(ShipNavigated(nav:, fuel:, events:))
      })
    _ -> parse_error_response(response)
  }
}

pub type ShipWarped {
  ShipWarped(nav: ShipNav, fuel: ShipFuel, events: List(ShipConditionEvent))
}

pub fn warp_ship_request(
  token: AgentToken,
  ship_symbol: ShipSymbol,
  waypoint_symbol: WaypointSymbol,
) -> Request(BitArray) {
  post_json(
    AgentAuth(token),
    "/my/ships/" <> ship_symbol.to_string(ship_symbol) <> "/warp",
    json.object([#("waypointSymbol", waypoint_symbol.encode(waypoint_symbol))]),
  )
}

pub fn warp_ship_response(
  response: Response(BitArray),
) -> ApiResponse(ShipWarped) {
  case response.status {
    200 ->
      parse_data_response(response, {
        use nav <- decode.field("nav", ship_nav.decoder())
        use fuel <- decode.field("fuel", ship_fuel.decoder())
        use events <- decode.field(
          "events",
          decode.list(ship_condition_event.decoder()),
        )
        decode.success(ShipWarped(nav:, fuel:, events:))
      })
    _ -> parse_error_response(response)
  }
}

pub fn orbit_ship_request(
  token: AgentToken,
  ship_symbol: ShipSymbol,
) -> Request(BitArray) {
  post(
    AgentAuth(token),
    "/my/ships/" <> ship_symbol.to_string(ship_symbol) <> "/orbit",
  )
}

pub fn orbit_ship_response(response: Response(BitArray)) -> ApiResponse(ShipNav) {
  case response.status {
    200 ->
      parse_data_response(
        response,
        decode.field("nav", ship_nav.decoder(), decode.success),
      )
    _ -> parse_error_response(response)
  }
}

pub type CargoPurchased {
  CargoPurchased(cargo: ShipCargo, transaction: MarketTransaction, agent: Agent)
}

pub fn purchase_cargo_request(
  token: AgentToken,
  ship_symbol: ShipSymbol,
  trade_symbol: TradeSymbol,
) -> Request(BitArray) {
  post_json(
    AgentAuth(token),
    "/my/ships/" <> ship_symbol.to_string(ship_symbol) <> "/purchase",
    json.object([#("symbol", trade_symbol.encode(trade_symbol))]),
  )
}

pub fn purchase_cargo_response(
  response: Response(BitArray),
) -> ApiResponse(CargoPurchased) {
  case response.status {
    201 ->
      parse_data_response(response, {
        use cargo <- decode.field("cargo", ship_cargo.decoder())
        use transaction <- decode.field(
          "transaction",
          market_transaction.decoder(),
        )
        use agent <- decode.field("agent", agent.decoder())
        decode.success(CargoPurchased(cargo:, transaction:, agent:))
      })
    _ -> parse_error_response(response)
  }
}

pub type ShipRefined {
  ShipRefined(
    cargo: ShipCargo,
    cooldown: Cooldown,
    produced: List(RefinementYield),
    consumed: List(RefinementYield),
  )
}

pub fn refine_ship_request(
  token: AgentToken,
  ship_symbol: ShipSymbol,
  produce: RefinementProduce,
) -> Request(BitArray) {
  post_json(
    AgentAuth(token),
    "/my/ships/" <> ship_symbol.to_string(ship_symbol) <> "/refine",
    json.object([#("produce", refinement_produce.encode(produce))]),
  )
}

pub fn refine_ship_response(
  response: Response(BitArray),
) -> ApiResponse(ShipRefined) {
  case response.status {
    201 ->
      parse_data_response(response, {
        use cargo <- decode.field("cargo", ship_cargo.decoder())
        use cooldown <- decode.field("cooldown", cooldown.decoder())
        use produced <- decode.field(
          "produced",
          decode.list(refinement_yield.decoder()),
        )
        use consumed <- decode.field(
          "consumed",
          decode.list(refinement_yield.decoder()),
        )
        decode.success(ShipRefined(cargo:, cooldown:, produced:, consumed:))
      })
    _ -> parse_error_response(response)
  }
}

pub type ShipRefueled {
  ShipRefueled(
    agent: Agent,
    fuel: ShipFuel,
    cargo: ShipCargo,
    transaction: MarketTransaction,
  )
}

pub fn refuel_ship_request(
  token: AgentToken,
  ship_symbol: ShipSymbol,
  units: Option(Int),
  from_cargo: Option(Bool),
) -> Request(BitArray) {
  post_json(
    AgentAuth(token),
    "/my/ships/" <> ship_symbol.to_string(ship_symbol) <> "/refuel",
    json.object([
      #("units", json.nullable(units, json.int)),
      #("fromCargo", json.nullable(from_cargo, json.bool)),
    ]),
  )
}

pub fn refuel_ship_response(
  response: Response(BitArray),
) -> ApiResponse(ShipRefueled) {
  case response.status {
    200 ->
      parse_data_response(response, {
        use agent <- decode.field("agent", agent.decoder())
        use fuel <- decode.field("fuel", ship_fuel.decoder())
        use cargo <- decode.field("cargo", ship_cargo.decoder())
        use transaction <- decode.field(
          "transaction",
          market_transaction.decoder(),
        )
        decode.success(ShipRefueled(agent:, fuel:, cargo:, transaction:))
      })
    _ -> parse_error_response(response)
  }
}

pub type ShipRepaired {
  ShipRepaired(agent: Agent, ship: Ship, transaction: RepairTransaction)
}

pub fn repair_ship_request(
  token: AgentToken,
  ship_symbol: ShipSymbol,
) -> Request(BitArray) {
  post(
    AgentAuth(token),
    "/my/ships/" <> ship_symbol.to_string(ship_symbol) <> "/repair",
  )
}

pub fn repair_ship_response(
  response: Response(BitArray),
) -> ApiResponse(ShipRepaired) {
  case response.status {
    200 ->
      parse_data_response(response, {
        use agent <- decode.field("agent", agent.decoder())
        use ship <- decode.field("ship", ship.decoder())
        use transaction <- decode.field(
          "transaction",
          repair_transaction.decoder(),
        )
        decode.success(ShipRepaired(agent:, ship:, transaction:))
      })
    _ -> parse_error_response(response)
  }
}

pub fn get_repair_ship_request(
  token: AgentToken,
  ship_symbol: ShipSymbol,
) -> Request(BitArray) {
  get(
    AgentAuth(token),
    "/my/ships/" <> ship_symbol.to_string(ship_symbol) <> "/repair",
  )
}

pub fn get_repair_ship_response(
  response: Response(BitArray),
) -> ApiResponse(RepairTransaction) {
  case response.status {
    200 ->
      parse_data_response(
        response,
        decode.field(
          "transaction",
          repair_transaction.decoder(),
          decode.success,
        ),
      )
    _ -> parse_error_response(response)
  }
}

pub type CargoSold {
  CargoSold(cargo: ShipCargo, transaction: MarketTransaction, agent: Agent)
}

pub fn sell_cargo_request(
  token: AgentToken,
  ship_symbol: ShipSymbol,
  trade_symbol: TradeSymbol,
  units: Int,
) -> Request(BitArray) {
  post_json(
    AgentAuth(token),
    "/my/ships/" <> ship_symbol.to_string(ship_symbol) <> "/sell",
    json.object([
      #("symbol", trade_symbol.encode(trade_symbol)),
      #("units", json.int(units)),
    ]),
  )
}

pub fn sell_cargo_response(
  response: Response(BitArray),
) -> ApiResponse(CargoSold) {
  case response.status {
    201 ->
      parse_data_response(response, {
        use cargo <- decode.field("cargo", ship_cargo.decoder())
        use transaction <- decode.field(
          "transaction",
          market_transaction.decoder(),
        )
        use agent <- decode.field("agent", agent.decoder())
        decode.success(CargoSold(cargo:, transaction:, agent:))
      })
    _ -> parse_error_response(response)
  }
}

pub type ResourcesSiphoned {
  ResourcesSiphoned(
    siphon: Siphon,
    cooldown: Cooldown,
    cargo: ShipCargo,
    events: List(ShipConditionEvent),
  )
}

pub fn siphon_resources_request(
  token: AgentToken,
  ship_symbol: ShipSymbol,
) -> Request(BitArray) {
  post(
    AgentAuth(token),
    "/my/ships/" <> ship_symbol.to_string(ship_symbol) <> "/siphon",
  )
}

pub fn siphon_resources_response(
  response: Response(BitArray),
) -> ApiResponse(ResourcesSiphoned) {
  case response.status {
    201 ->
      parse_data_response(response, {
        use siphon <- decode.field("siphon", siphon.decoder())
        use cooldown <- decode.field("cooldown", cooldown.decoder())
        use cargo <- decode.field("cargo", ship_cargo.decoder())
        use events <- decode.field(
          "events",
          decode.list(ship_condition_event.decoder()),
        )
        decode.success(ResourcesSiphoned(siphon:, cooldown:, cargo:, events:))
      })
    _ -> parse_error_response(response)
  }
}

pub type SurveyCreated {
  SurveyCreated(cooldown: Cooldown, surveys: List(Survey))
}

pub fn create_survey_request(
  token: AgentToken,
  ship_symbol: ShipSymbol,
) -> Request(BitArray) {
  post(
    AgentAuth(token),
    "/my/ships/" <> ship_symbol.to_string(ship_symbol) <> "/survey",
  )
}

pub fn create_survey_response(
  response: Response(BitArray),
) -> ApiResponse(SurveyCreated) {
  case response.status {
    201 ->
      parse_data_response(response, {
        use cooldown <- decode.field("cooldown", cooldown.decoder())
        use surveys <- decode.field("surveys", decode.list(survey.decoder()))
        decode.success(SurveyCreated(cooldown:, surveys:))
      })
    _ -> parse_error_response(response)
  }
}

pub type CargoTransferred {
  CargoTransferred(cargo: ShipCargo, target_cargo: ShipCargo)
}

pub fn transfer_cargo_request(
  token: AgentToken,
  ship_symbol: ShipSymbol,
  trade_symbol: TradeSymbol,
  units: Int,
  target_ship_symbol: ShipSymbol,
) -> Request(BitArray) {
  post_json(
    AgentAuth(token),
    "/my/ships/" <> ship_symbol.to_string(ship_symbol) <> "/transfer",
    json.object([
      #("tradeSymbol", trade_symbol.encode(trade_symbol)),
      #("units", json.int(units)),
      #("shipSymbol", ship_symbol.encode(target_ship_symbol)),
    ]),
  )
}

pub fn transfer_cargo_response(
  response: Response(BitArray),
) -> ApiResponse(CargoTransferred) {
  case response.status {
    200 ->
      parse_data_response(response, {
        use cargo <- decode.field("cargo", ship_cargo.decoder())
        use target_cargo <- decode.field("targetCargo", ship_cargo.decoder())
        decode.success(CargoTransferred(cargo:, target_cargo:))
      })
    _ -> parse_error_response(response)
  }
}

pub fn get_ship_cargo_request(
  token: AgentToken,
  ship_symbol: ShipSymbol,
) -> Request(BitArray) {
  get(
    AgentAuth(token),
    "/my/ships/" <> ship_symbol.to_string(ship_symbol) <> "/cargo",
  )
}

pub fn get_ship_cargo_response(
  response: Response(BitArray),
) -> ApiResponse(ShipCargo) {
  case response.status {
    200 -> parse_data_response(response, ship_cargo.decoder())
    _ -> parse_error_response(response)
  }
}

pub fn get_ship_modules_request(
  token: AgentToken,
  ship_symbol: ShipSymbol,
) -> Request(BitArray) {
  get(
    AgentAuth(token),
    "/my/ships/" <> ship_symbol.to_string(ship_symbol) <> "/modules",
  )
}

pub fn get_ship_modules_response(
  response: Response(BitArray),
) -> ApiResponse(List(ShipModule)) {
  case response.status {
    200 -> parse_data_response(response, decode.list(ship_module.decoder()))
    _ -> parse_error_response(response)
  }
}

pub type ShipModuleInstalled {
  ShipModuleInstalled(
    agent: Agent,
    modules: List(ShipModule),
    cargo: ShipCargo,
    transaction: ShipModificationTransaction,
  )
}

pub fn install_ship_module_request(
  token: AgentToken,
  ship_symbol: ShipSymbol,
  module_symbol: ModuleSymbol,
) -> Request(BitArray) {
  post_json(
    AgentAuth(token),
    "/my/ships/" <> ship_symbol.to_string(ship_symbol) <> "/modules/install",
    json.object([#("symbol", module_symbol.encode(module_symbol))]),
  )
}

pub fn install_ship_module_response(
  response: Response(BitArray),
) -> ApiResponse(ShipModuleInstalled) {
  case response.status {
    201 ->
      parse_data_response(response, {
        use agent <- decode.field("agent", agent.decoder())
        use modules <- decode.field(
          "modules",
          decode.list(ship_module.decoder()),
        )
        use cargo <- decode.field("cargo", ship_cargo.decoder())
        use transaction <- decode.field(
          "transaction",
          ship_modification_transaction.decoder(),
        )
        decode.success(ShipModuleInstalled(
          agent:,
          modules:,
          cargo:,
          transaction:,
        ))
      })
    _ -> parse_error_response(response)
  }
}

pub type ShipModuleRemoved {
  ShipModuleRemoved(
    agent: Agent,
    modules: List(ShipModule),
    cargo: ShipCargo,
    transaction: ShipModificationTransaction,
  )
}

pub fn remove_ship_module_request(
  token: AgentToken,
  ship_symbol: ShipSymbol,
  module_symbol: ModuleSymbol,
) -> Request(BitArray) {
  post_json(
    AgentAuth(token),
    "/my/ships/" <> ship_symbol.to_string(ship_symbol) <> "/modules/remove",
    json.object([#("symbol", module_symbol.encode(module_symbol))]),
  )
}

pub fn remove_ship_module_response(
  response: Response(BitArray),
) -> ApiResponse(ShipModuleRemoved) {
  case response.status {
    201 ->
      parse_data_response(response, {
        use agent <- decode.field("agent", agent.decoder())
        use modules <- decode.field(
          "modules",
          decode.list(ship_module.decoder()),
        )
        use cargo <- decode.field("cargo", ship_cargo.decoder())
        use transaction <- decode.field(
          "transaction",
          ship_modification_transaction.decoder(),
        )
        decode.success(ShipModuleRemoved(agent:, modules:, cargo:, transaction:))
      })
    _ -> parse_error_response(response)
  }
}

pub fn get_ship_mounts_request(
  token: AgentToken,
  ship_symbol: ShipSymbol,
) -> Request(BitArray) {
  get(
    AgentAuth(token),
    "/my/ships/" <> ship_symbol.to_string(ship_symbol) <> "/mounts",
  )
}

pub fn get_ship_mounts_response(
  response: Response(BitArray),
) -> ApiResponse(List(ShipMount)) {
  case response.status {
    200 -> parse_data_response(response, decode.list(ship_mount.decoder()))
    _ -> parse_error_response(response)
  }
}

pub type ShipMountInstalled {
  ShipMountInstalled(
    agent: Agent,
    mounts: List(ShipMount),
    cargo: ShipCargo,
    transaction: ShipModificationTransaction,
  )
}

pub fn install_ship_mount_request(
  token: AgentToken,
  ship_symbol: ShipSymbol,
  mount_symbol: MountSymbol,
) -> Request(BitArray) {
  post_json(
    AgentAuth(token),
    "/my/ships/" <> ship_symbol.to_string(ship_symbol) <> "/mounts/install",
    json.object([#("symbol", mount_symbol.encode(mount_symbol))]),
  )
}

pub fn install_ship_mount_response(
  response: Response(BitArray),
) -> ApiResponse(ShipMountInstalled) {
  case response.status {
    201 ->
      parse_data_response(response, {
        use agent <- decode.field("agent", agent.decoder())
        use mounts <- decode.field("mounts", decode.list(ship_mount.decoder()))
        use cargo <- decode.field("cargo", ship_cargo.decoder())
        use transaction <- decode.field(
          "transaction",
          ship_modification_transaction.decoder(),
        )
        decode.success(ShipMountInstalled(agent:, mounts:, cargo:, transaction:))
      })
    _ -> parse_error_response(response)
  }
}

pub type ShipMountRemoved {
  ShipMountRemoved(
    agent: Agent,
    mounts: List(ShipMount),
    cargo: ShipCargo,
    transaction: ShipModificationTransaction,
  )
}

pub fn remove_ship_mount_request(
  token: AgentToken,
  ship_symbol: ShipSymbol,
  mount_symbol: MountSymbol,
) -> Request(BitArray) {
  post_json(
    AgentAuth(token),
    "/my/ships/" <> ship_symbol.to_string(ship_symbol) <> "/mounts/remove",
    json.object([#("symbol", mount_symbol.encode(mount_symbol))]),
  )
}

pub fn remove_ship_mount_response(
  response: Response(BitArray),
) -> ApiResponse(ShipMountRemoved) {
  case response.status {
    201 ->
      parse_data_response(response, {
        use agent <- decode.field("agent", agent.decoder())
        use mounts <- decode.field("mounts", decode.list(ship_mount.decoder()))
        use cargo <- decode.field("cargo", ship_cargo.decoder())
        use transaction <- decode.field(
          "transaction",
          ship_modification_transaction.decoder(),
        )
        decode.success(ShipMountRemoved(agent:, mounts:, cargo:, transaction:))
      })
    _ -> parse_error_response(response)
  }
}

pub fn get_ship_nav_request(
  token: AgentToken,
  ship_symbol: ShipSymbol,
) -> Request(BitArray) {
  get(
    AgentAuth(token),
    "/my/ships/" <> ship_symbol.to_string(ship_symbol) <> "/nav",
  )
}

pub fn get_ship_nav_response(
  response: Response(BitArray),
) -> ApiResponse(ShipNav) {
  case response.status {
    200 -> parse_data_response(response, ship_nav.decoder())
    _ -> parse_error_response(response)
  }
}

pub type ShipNavPatched {
  ShipNavPatched(nav: ShipNav, fuel: ShipFuel, events: List(ShipConditionEvent))
}

pub fn patch_ship_nav_request(
  token: AgentToken,
  ship_symbol: ShipSymbol,
  flight_mode: ShipNavFlightMode,
) -> Request(BitArray) {
  patch_json(
    AgentAuth(token),
    "/my/ships/" <> ship_symbol.to_string(ship_symbol) <> "/nav",
    json.object([#("flightMode", ship_nav_flight_mode.encode(flight_mode))]),
  )
}

pub fn patch_ship_nav_response(
  response: Response(BitArray),
) -> ApiResponse(ShipNavPatched) {
  case response.status {
    200 ->
      parse_data_response(response, {
        use nav <- decode.field("nav", ship_nav.decoder())
        use fuel <- decode.field("fuel", ship_fuel.decoder())
        use events <- decode.field(
          "events",
          decode.list(ship_condition_event.decoder()),
        )
        decode.success(ShipNavPatched(nav:, fuel:, events:))
      })
    _ -> parse_error_response(response)
  }
}

pub type Stats {
  Stats(
    accounts: Option(Int),
    agents: Int,
    ships: Int,
    systems: Int,
    waypoints: Int,
  )
}

fn stats_decoder() -> Decoder(Stats) {
  use accounts <- decode.optional_field(
    "accounts",
    option.None,
    decode.optional(decode.int),
  )
  use agents <- decode.field("agents", decode.int)
  use ships <- decode.field("ships", decode.int)
  use systems <- decode.field("systems", decode.int)
  use waypoints <- decode.field("waypoints", decode.int)
  decode.success(Stats(accounts:, agents:, ships:, systems:, waypoints:))
}

pub type Health {
  Health(last_market_update: Option(Timestamp))
}

fn health_decoder() -> Decoder(Health) {
  use last_market_update <- decode.optional_field(
    "lastMarketUpdate",
    option.None,
    decode.optional(timestamp.decoder()),
  )
  decode.success(Health(last_market_update:))
}

pub type CreditLeaderboardEntry {
  CreditLeaderboardEntry(agent_symbol: AgentSymbol, credits: Int)
}

fn credit_leaderboard_entry_decoder() -> Decoder(CreditLeaderboardEntry) {
  use agent_symbol <- decode.field("agentSymbol", agent_symbol.decoder())
  use credits <- decode.field("credits", decode.int)
  decode.success(CreditLeaderboardEntry(agent_symbol:, credits:))
}

pub type ChartLeaderboardEntry {
  ChartLeaderboardEntry(agent_symbol: AgentSymbol, chart_count: Int)
}

fn chart_leaderboard_entry_decoder() -> Decoder(ChartLeaderboardEntry) {
  use agent_symbol <- decode.field("agentSymbol", agent_symbol.decoder())
  use chart_count <- decode.field("chartCount", decode.int)
  decode.success(ChartLeaderboardEntry(agent_symbol:, chart_count:))
}

pub type Leaderboards {
  Leaderboards(
    most_credits: List(CreditLeaderboardEntry),
    most_submitted_charts: List(ChartLeaderboardEntry),
  )
}

fn leaderboards_decoder() -> Decoder(Leaderboards) {
  use most_credits <- decode.field(
    "mostCredits",
    decode.list(credit_leaderboard_entry_decoder()),
  )
  use most_submitted_charts <- decode.field(
    "mostSubmittedCharts",
    decode.list(chart_leaderboard_entry_decoder()),
  )
  decode.success(Leaderboards(most_credits:, most_submitted_charts:))
}

pub type ServerResets {
  ServerResets(next: Timestamp, frequency: String)
}

fn server_resets_decoder() -> Decoder(ServerResets) {
  use next <- decode.field("next", timestamp.decoder())
  use frequency <- decode.field("frequency", decode.string)
  decode.success(ServerResets(next:, frequency:))
}

pub type Announcement {
  Announcement(title: String, body: String)
}

fn announcement_decoder() -> Decoder(Announcement) {
  use title <- decode.field("title", decode.string)
  use body <- decode.field("body", decode.string)
  decode.success(Announcement(title:, body:))
}

pub type Link {
  Link(name: String, url: Uri)
}

fn link_decoder() -> Decoder(Link) {
  use name <- decode.field("name", decode.string)
  use url <- decode.field("url", {
    use value <- decode.then(decode.string)
    case uri.parse(value) {
      Ok(time) -> decode.success(time)
      Error(Nil) -> decode.failure(uri.empty, "Uri")
    }
  })
  decode.success(Link(name:, url:))
}

pub type ServerStatus {
  ServerStatus(
    status: String,
    version: String,
    reset_date: Date,
    description: String,
    stats: Stats,
    health: Health,
    leaderboards: Leaderboards,
    server_resets: ServerResets,
    announcements: List(Announcement),
    links: List(Link),
  )
}

fn server_status_decoder() -> Decoder(ServerStatus) {
  use status <- decode.field("status", decode.string)
  use version <- decode.field("version", decode.string)
  use reset_date <- decode.field("resetDate", date.decoder())
  use description <- decode.field("description", decode.string)
  use stats <- decode.field("stats", stats_decoder())
  use health <- decode.field("health", health_decoder())
  use leaderboards <- decode.field("leaderboards", leaderboards_decoder())
  use server_resets <- decode.field("serverResets", server_resets_decoder())
  use announcements <- decode.field(
    "announcements",
    decode.list(announcement_decoder()),
  )
  use links <- decode.field("links", decode.list(link_decoder()))
  decode.success(ServerStatus(
    status:,
    version:,
    reset_date:,
    description:,
    stats:,
    health:,
    leaderboards:,
    server_resets:,
    announcements:,
    links:,
  ))
}

pub fn get_server_status_request() -> Request(BitArray) {
  get(NoAuth, "/")
}

pub fn get_server_status_response(
  response: Response(BitArray),
) -> ApiResponse(ServerStatus) {
  case response.status {
    200 -> parse_response(response, server_status_decoder())
    _ -> parse_error_response(response)
  }
}

pub type ErrorCode {
  ErrorCode(code: Int, name: String)
}

fn error_code_decoder() -> Decoder(ErrorCode) {
  use code <- decode.field("code", decode.int)
  use name <- decode.field("name", decode.string)
  decode.success(ErrorCode(code:, name:))
}

pub fn list_error_codes_request() -> Request(BitArray) {
  get(NoAuth, "/error-codes")
}

pub fn list_error_codes_response(
  response: Response(BitArray),
) -> ApiResponse(List(ErrorCode)) {
  case response.status {
    200 ->
      parse_response(
        response,
        decode.field(
          "errorCodes",
          decode.list(error_code_decoder()),
          decode.success,
        ),
      )
    _ -> parse_error_response(response)
  }
}

pub fn list_systems_request(
  page: Option(Int),
  limit: Option(Int),
) -> Request(BitArray) {
  get_page(NoAuth, "/systems", page, limit)
}

pub fn list_systems_response(
  response: Response(BitArray),
) -> ApiResponse(PagedData(List(System))) {
  case response.status {
    200 -> parse_paged_data_response(response, decode.list(system.decoder()))
    _ -> parse_error_response(response)
  }
}

pub fn get_system_request(system_symbol: SystemSymbol) -> Request(BitArray) {
  get(NoAuth, "/systems/" <> system_symbol.to_string(system_symbol))
}

pub fn get_system_response(response: Response(BitArray)) -> ApiResponse(System) {
  case response.status {
    200 -> parse_data_response(response, system.decoder())
    _ -> parse_error_response(response)
  }
}

pub fn list_system_waypoints_request(
  system_symbol: SystemSymbol,
  page: Option(Int),
  limit: Option(Int),
  type_: Option(WaypointType),
  traits: List(WaypointTraitSymbol),
) -> Request(BitArray) {
  let query =
    list.map(traits, fn(trait) {
      #("traits", waypoint_trait_symbol.to_string(trait))
    })
  let query = case type_ {
    option.Some(type_) -> [#("type", waypoint_type.to_string(type_)), ..query]
    option.None -> query
  }
  get_page_with_query(
    NoAuth,
    "/systems/" <> system_symbol.to_string(system_symbol) <> "/waypoints",
    page,
    limit,
    query,
  )
}

pub fn list_system_waypoints_response(
  response: Response(BitArray),
) -> ApiResponse(PagedData(List(Waypoint))) {
  case response.status {
    200 -> parse_paged_data_response(response, decode.list(waypoint.decoder()))
    _ -> parse_error_response(response)
  }
}

pub fn get_waypoint_request(
  system_symbol: SystemSymbol,
  waypoint_symbol: WaypointSymbol,
) -> Request(BitArray) {
  get(
    NoAuth,
    "/systems/"
      <> system_symbol.to_string(system_symbol)
      <> "/waypoints/"
      <> waypoint_symbol.to_string(waypoint_symbol),
  )
}

pub fn get_waypoint_response(
  response: Response(BitArray),
) -> ApiResponse(Waypoint) {
  case response.status {
    200 -> parse_data_response(response, waypoint.decoder())
    _ -> parse_error_response(response)
  }
}

pub fn get_construction_site_request(
  system_symbol: SystemSymbol,
  waypoint_symbol: WaypointSymbol,
) -> Request(BitArray) {
  get(
    NoAuth,
    "/systems/"
      <> system_symbol.to_string(system_symbol)
      <> "/waypoints/"
      <> waypoint_symbol.to_string(waypoint_symbol)
      <> "/construction",
  )
}

pub fn get_construction_site_response(
  response: Response(BitArray),
) -> ApiResponse(Construction) {
  case response.status {
    200 -> parse_data_response(response, construction.decoder())
    _ -> parse_error_response(response)
  }
}

pub type ConstructionSiteSupplied {
  ConstructionSiteSupplied(construction: Construction, cargo: ShipCargo)
}

pub fn supply_construction_site_request(
  token: AgentToken,
  system_symbol: SystemSymbol,
  waypoint_symbol: WaypointSymbol,
  ship_symbol: ShipSymbol,
  trade_symbol: TradeSymbol,
  units: Int,
) -> Request(BitArray) {
  post_json(
    AgentAuth(token),
    "/systems/"
      <> system_symbol.to_string(system_symbol)
      <> "/waypoints/"
      <> waypoint_symbol.to_string(waypoint_symbol)
      <> "/construction/supply",
    json.object([
      #("shipSymbol", ship_symbol.encode(ship_symbol)),
      #("tradeSymbol", trade_symbol.encode(trade_symbol)),
      #("units", json.int(units)),
    ]),
  )
}

pub fn supply_construction_site_response(
  response: Response(BitArray),
) -> ApiResponse(ConstructionSiteSupplied) {
  case response.status {
    201 ->
      parse_data_response(response, {
        use construction <- decode.field("construction", construction.decoder())
        use cargo <- decode.field("cargo", ship_cargo.decoder())
        decode.success(ConstructionSiteSupplied(construction:, cargo:))
      })
    _ -> parse_error_response(response)
  }
}

pub fn get_market_request(
  system_symbol: SystemSymbol,
  waypoint_symbol: WaypointSymbol,
) -> Request(BitArray) {
  get(
    NoAuth,
    "/systems/"
      <> system_symbol.to_string(system_symbol)
      <> "/waypoints/"
      <> waypoint_symbol.to_string(waypoint_symbol)
      <> "/market",
  )
}

pub fn get_market_response(response: Response(BitArray)) -> ApiResponse(Market) {
  case response.status {
    200 -> parse_data_response(response, market.decoder())
    _ -> parse_error_response(response)
  }
}

pub fn get_jump_gate_request(
  system_symbol: SystemSymbol,
  waypoint_symbol: WaypointSymbol,
) -> Request(BitArray) {
  get(
    NoAuth,
    "/systems/"
      <> system_symbol.to_string(system_symbol)
      <> "/waypoints/"
      <> waypoint_symbol.to_string(waypoint_symbol)
      <> "/jump-gate",
  )
}

pub fn get_jump_gate_response(
  response: Response(BitArray),
) -> ApiResponse(JumpGate) {
  case response.status {
    200 -> parse_data_response(response, jump_gate.decoder())
    _ -> parse_error_response(response)
  }
}

pub fn get_shipyard_request(
  system_symbol: SystemSymbol,
  waypoint_symbol: WaypointSymbol,
) -> Request(BitArray) {
  get(
    NoAuth,
    "/systems/"
      <> system_symbol.to_string(system_symbol)
      <> "/waypoints/"
      <> waypoint_symbol.to_string(waypoint_symbol)
      <> "/shipyard",
  )
}

pub fn get_shipyard_response(
  response: Response(BitArray),
) -> ApiResponse(Shipyard) {
  case response.status {
    200 -> parse_data_response(response, shipyard.decoder())
    _ -> parse_error_response(response)
  }
}
